///////////////////////////////////////////////////////////////////
/*
 *	ACR92 USB Smart Card Reader Driver Kernel module
 *  Copyright (C) 2005-2006  TL Sudheendran(tlsudheendran@scmmicro.co.in)

 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.

 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.

 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


 *	SCM Microsystems  Inc., hereby disclaims all copyright interest in the
 *	library acr92.o  written by TL Sudheendran


	Module		:	acr92_main.c

	Author		:	SCM Microsystems

	Development History :

		Date		Version No.		Changes
		_______________________________________________
		02 Feb, 2004	2.0.1		Initial Versions
		30 Mar, 2004	3.0.1		Added support for suspend, resume
									under event handler.
		22 Nov, 2004	3.1.1		TTP issues 1359,1360
		25 Feb, 2005	3.2.1		Added support for Linux 2.6.x; TTP 1410, 2139
		17 May, 2005	4.1.1		Removed known limitations regarding surprise
									removal; Added LGPL License agreement

		20 Dec,	2005	4.1.2		Added support for kernel version 2.6.13
		30 Jan,	2006	4.2.1		Dec 2005 fix being released properly.

		10 Aug, 2006	4.2.2		Added support for kernel 2.6.16
		04 Oct, 2006	4.2.3		Added support for kernel versions upto 2.6.18
		28 Mar, 2007	4.2.4		Added support upto 2.6.20
		15 Apr, 2009	4.2.6		Added support upto 2.6.29
*/



#include "includes.h"

#define CS_CHECK(fn, args...) \
if(fn(args)) goto cs_failed

#define CFG_CHECK(fn, args...) \
if ((fn(args)) != 0) goto next_entry

#define 		MAX_PCMCIA_DEVICES 	8
#define			FIRST_INSTANCE		0

DEVICE_EXTENSION	sDeviceExtension[MAX_PCMCIA_DEVICES];

dev_t	dev;
struct	cdev	*ACR92_cdev;

// Data type changed by jithu in Version 4.6.3.
// to accomodate changes in kernel
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,17)
	struct pcmcia_device *dev_list 		= NULL;
#else
	dev_link_t 	*dev_list 		= NULL;
#endif

dev_info_t 	dev_info 		= "acr92_cs";
int			free_ports 		= 0;
u_int 		irq_mask 		= 0xdeb8;
int 		irq_list[4]		= { -1 };
BOOL		gWaitEvent		= FALSE;
int 		nInstanceCount 		= 0;
int 		nMajorNumber		= 0;
BOOL 		bMajorNumberAssigned	= FALSE;

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,11)
	static struct pcmcia_device_id acr92_ids[] =
	{
		PCMCIA_DEVICE_PROD_ID1("ACR92 PCMCIA",0xfb55d54c),
		PCMCIA_DEVICE_NULL
	};
	MODULE_DEVICE_TABLE(pcmcia, acr92_ids);

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
		static struct pcmcia_driver acr92_driver = {

			.owner		= THIS_MODULE,
			.drv		= {
							.name	="acr92_cs",
						  },
			.probe		= Driver_Attach,
			.remove		= Driver_Detach,
			.suspend	= Driver_Suspend,
			.resume		= Driver_Resume,

			.id_table	= acr92_ids,
		};
	#else
		static struct pcmcia_driver acr92_driver = {

		.owner		= THIS_MODULE,
		.drv		= {
						.name	="acr92_cs",
					  },
		.attach		= Driver_Attach,
		.event		= Event_Handler,
		.detach		= Driver_Detach,
		.id_table	= acr92_ids,
	};
	#endif

#else
static struct pcmcia_driver acr92_driver = {

        .owner      = THIS_MODULE,
        .drv        = {
                        .name   ="acr92_cs",
                      },
        .attach     = Driver_Attach,
        .detach     = Driver_Detach,
    };
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,17)
	static void Driver_Config( struct pcmcia_device *link );	
#else
	static void Driver_Config( dev_link_t *link );
#endif

struct file_operations ACR92_fops =
{
	read: 	IFDHRead,
	ioctl: 	IFDHIoctl,
	write: 	IFDHWrite,
	open: 	IFDHOpen,
	release:IFDHClose,
};

static int __init init_driver( void )
{
	int nRetVal=1;

	DebugLog ("\n________(** Calling Init Driver of ACR92 Module now **)__________ \n");

	nRetVal = alloc_chrdev_region (&dev, FIRST_INSTANCE, MAX_PCMCIA_DEVICES, "ACR92");
	nMajorNumber = MAJOR (dev);

	if (nRetVal < 0)
	{
		DebugLog (" *** init_driver : Cannot get major number! ***\n");
		return nRetVal;
	}

	nRetVal = pcmcia_register_driver(&acr92_driver);

	ACR92_cdev = cdev_alloc();
	ACR92_cdev->ops = &ACR92_fops;
	ACR92_cdev->owner = THIS_MODULE;
	nRetVal = cdev_add (ACR92_cdev, dev, MAX_PCMCIA_DEVICES);

	if ( nRetVal  )
	{
		DebugLog ( "init_driver : cdev_add failed");
	}

	DebugLog (" Calling Init Driver of ACR92 Module EXIT %d\n", nRetVal );

	return nRetVal;
}

static void __exit exit_driver(void)
{
	pcmcia_unregister_driver(&acr92_driver);

	DebugLog ("\n Calling Exit Driver of ACR92 Module\n");

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
	while (dev_list != NULL)
	{
		if (dev_list->state & DEV_CONFIG)
		{
			Driver_Release ((u_long)dev_list);
		}
 		Driver_Detach(dev_list);
	}
#endif

	cdev_del (ACR92_cdev);

	dev = MKDEV (nMajorNumber, FIRST_INSTANCE);
	unregister_chrdev_region (dev, MAX_PCMCIA_DEVICES);

	DebugLog (" Calling Exit Driver of ACR92 Module EXIT\n");
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,17)

int Driver_Attach (struct pcmcia_device *p_dev)
{
	int		nCount;

	DEVICE_EXTENSION	*pDeviceExtension = NULL;

	DebugLog ("Inside Driver_Attach ----- \n");

	pDeviceExtension = &sDeviceExtension [nInstanceCount];
	RtlZeroMemory (pDeviceExtension, sizeof(DEVICE_EXTENSION));

	pDeviceExtension->pucDeviceName = (PUCHAR ) kmalloc( sizeof(UCHAR) * 32, GFP_KERNEL );
	if( NULL == pDeviceExtension->pucDeviceName )
	{
		pDeviceExtension = NULL;
		DebugLog ("Driver_Attach : kmalloc failed for pucDeviceName\n");
		return -ENOMEM;
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
	INIT_WORK(&pDeviceExtension->stWorkQueue, tqGetCardStatus);
#endif	

	pDeviceExtension->DeviceLink = p_dev;

	pDeviceExtension->CardTracker.pDeviceExtension = pDeviceExtension;


	// End of Allocation of memories
	// setting default values to device extension structure.
	pDeviceExtension->nMajorNumber 		= -1;
	pDeviceExtension->nInstanceCount 	= nInstanceCount;
	pDeviceExtension->bBootRomPresent	= FALSE;
	pDeviceExtension->bIsPolledMode 	= FALSE;

	// pcsc daemon does not issue a power request to update
	// card status on system resume from standby/suspend.
	// hence, a card absent/present is simuated from the
	// kernel module.
	pDeviceExtension->bIsSystemResume	= FALSE;

	sprintf (pDeviceExtension->pucDeviceName,"ACR92%d", nInstanceCount);

//	DebugLog (" Driver_Attach of ACR92 Module:: [%s]\n", pDeviceExtension->pucDeviceName);
	pDeviceExtension->nMajorNumber = nMajorNumber;

 	init_waitqueue_head ( &pDeviceExtension->tqsEventWait );
	init_waitqueue_head ( &pDeviceExtension->tqsReadEventWait );

	// Interrupt setup
	p_dev->irq.Attributes = IRQ_TYPE_DYNAMIC_SHARING | IRQ_HANDLE_PRESENT;
	p_dev->irq.IRQInfo1 = IRQ_INFO2_VALID | IRQ_LEVEL_ID;
	if (irq_list[0] == -1)
	{
		p_dev->irq.IRQInfo2 = irq_mask;
	}
	else
	{
		for (nCount = 0; nCount < 4; nCount++)
		{
			p_dev->irq.IRQInfo2 |= 1 << irq_list[nCount];
		}
	}

	p_dev->irq.Handler = &Driver_IrqHandler;
	p_dev->irq.Instance = pDeviceExtension;

	p_dev->conf.Attributes = CONF_ENABLE_IRQ;
	p_dev->conf.IntType = INT_MEMORY_AND_IO;

	Driver_Config(p_dev);

	DebugLog ("Driver_Attach EXIT \n");

	return 0;
}

void Driver_Detach(struct pcmcia_device *p_dev)
{
	int i;	
	DEVICE_EXTENSION *pDeviceExtension = NULL;

	DebugLog( "\n ACR92::Driver_Detach Entry \n");

	if(!p_dev)
	{
		DebugLog (" NULL pointer for p_dev\n");
		return;
	}

	for (i=0; i<MAX_PCMCIA_DEVICES; i++)
	{
		if ( p_dev == sDeviceExtension[i].DeviceLink )
		{
			pDeviceExtension = &sDeviceExtension[i];
			break;
		}
	}

	if ( !pDeviceExtension )
	{
		DebugLog (" Could not find device extension\n");
	}


	// Card Removal event moved up from event handler to here
	DebugLog("<<<------- card removed -------->>>\n");

	pDeviceExtension->nStop = 1;
	flush_scheduled_work();


	DebugLog ("\n Driver_Release\n");

	pcmcia_disable_device(p_dev);

	if( NULL != pDeviceExtension )
	{
		pDeviceExtension->bSignalEvent = TRUE;
		pDeviceExtension->bIsEventSet  = TRUE;
		wake_up_interruptible ( &pDeviceExtension->tqsReadEventWait );
	}
	else
	{
		DebugLog (" Driver_Detach : NULL pointer for pDeviceExtension\n");
		return;
	}

	// 	 Kill the card tracking thread
	if( TRUE == pDeviceExtension->bIsPolledMode )
	{
		DebugLog(" ACR921::Stopping the Polled Mode Card Tracking Thread.\n");
		pDeviceExtension->bIsPolledMode = FALSE;
		pDeviceExtension->bIsIrqSet = FALSE;

		send_sig(SIGTERM, pDeviceExtension->pDE_thread, 1);
		pDeviceExtension->pDE_thread = 0;
	}


	if( p_dev->priv )
	{
		kfree(p_dev->priv);
	}

	if( pDeviceExtension->pucDeviceName )
	{
		kfree( pDeviceExtension->pucDeviceName );
		pDeviceExtension->pucDeviceName = NULL;
	}

	if( pDeviceExtension )
	{
		pDeviceExtension = NULL;
	}

	DebugLog(" Driver_Detach Exit\n");
}

int Driver_Suspend(struct pcmcia_device *p_dev)
{
	int i;
	DEVICE_EXTENSION *pDeviceExtension = NULL;

	DebugLog( "\n ACR92::Driver_Suspend Entry \n");


	if(!p_dev)
	{
		DebugLog (" NULL pointer for pDeviceLink\n");
		return -1;
	}

	for (i=0; i<MAX_PCMCIA_DEVICES; i++)
	{
		if ( p_dev == sDeviceExtension[i].DeviceLink )
		{
			pDeviceExtension = &sDeviceExtension[i];
			break;
		}
	}

	if( NULL == pDeviceExtension )
	{
		DebugLog (" Driver_Suspend : NULL pointer for pDeviceExtension\n");
		return -1;
	}

	DebugLog("<<<------- card suspend-------->>>\n");
	// Fall through...
	
	// Mark the device as stopped, to block IO until later
	pDeviceExtension->nStop = 1;
	// from 2.6.17 release configuration will be called by the PCMCIA Core
	
	return 0;
}


int Driver_Resume(struct pcmcia_device *p_dev)
{
	int i;
	DEVICE_EXTENSION *pDeviceExtension = NULL;
	int ncsRetVal=0;

	DebugLog( "\n ACR92::Driver_Resume Entry \n");

	if(!p_dev)
	{
		DebugLog (" NULL pointer for pDeviceLink\n");
		return -1;
	}

	for (i=0; i<MAX_PCMCIA_DEVICES; i++)
	{
		if ( p_dev == sDeviceExtension[i].DeviceLink )
		{
			pDeviceExtension = &sDeviceExtension[i];
			break;
		}
	}

	if( NULL == pDeviceExtension )
	{
		DebugLog (" Driver_Resume : NULL pointer for pDeviceExtension\n");
		return -1;
	}

	DebugLog("<<<------- card resume-------->>>\n");
	DebugLog(" Event PM_RESUME Called \n");

	if(FALSE == pDeviceExtension->bIsIrqSet  ||TRUE == pDeviceExtension->bIsPolledMode)
	{
		DebugLog("******* THREAD TO BE STARTED ON RESUME*****\n");
		init_waitqueue_head(&pDeviceExtension->tqsCardStatusWait);
		pDeviceExtension->DE_thread_pid = kernel_thread(PcdmCardStatusThread, pDeviceExtension, CLONE_KERNEL);
	}
	
	DebugLog("<<<------- card reset-------->>>\n");

	// since the device went to standby/suspend, the firmware
	// will get reset to Rev A status. Hence, we need to update
	// the transport mechanism used prior to suspend/standby to the firmware.


		if( PCDM_REV_C == pDeviceExtension->byPcdmRev )
		{
			DebugLog("*********** Found REVC ON RESUME***************\n");
			// we set the transport mechanism to REv-A since the mode will 
			// now be A and kernel module shud also be called by A only.
			pDeviceExtension->byPcdmRev = PCDM_REV_A;
			ncsRetVal = IFDSetTransportMechanism( pDeviceExtension );
			if( 0 == ncsRetVal )
			{
				DebugLog (" Setting PCDM revision to REV C\n");
				pDeviceExtension->byPcdmRev = PCDM_REV_C;
			}
			else
			{
				// we failed to the tranport mechanism. hence, we revert
				// back to rev-A. we dont take into account the return
				// state
				DebugLog (" Setting PCDM revision to REV A\n");
				pDeviceExtension->byPcdmRev = PCDM_REV_A;
				ncsRetVal = IFDSetTransportMechanism( pDeviceExtension );
				DebugLog("Failure in Setting transport Mechanism to Rev-C. Revert to Rev-A\n");
				//break;
			}
		}

		DebugLog (" Event Handler [CS_EVENT_CARD_RESET] : IFDGetCardStatus\n");
		IFDGetCardStatus(pDeviceExtension);
		pDeviceExtension->bIsSystemResume = TRUE;
	
	//In a normal driver, additional code may go here to restore
	//the device state and restart IO.
	pDeviceExtension->nStop = 0;
	return 0;
}

int	ConfigCheck (struct pcmcia_device *link, cistpl_cftable_entry_t *cfg, \
	cistpl_cftable_entry_t *dflt, unsigned int vcc, void *priv_data)
{
	int nRetVal = 1;
	win_req_t *req = (win_req_t*) priv_data;

	do
	{
		if (cfg->index == 0)
		{
			break;
		}

		link->conf.ConfigIndex = cfg->index;

		// Does this card need audio output?
		if (cfg->flags & CISTPL_CFTABLE_AUDIO)
		{
			link->conf.Attributes |= CONF_ENABLE_SPKR;
			link->conf.Status = CCSR_AUDIO_ENA;
		}

		if (cfg->vpp1.present & (1<<CISTPL_POWER_VNOM))
		{
			link->conf.Vpp =
				cfg->vpp1.param[CISTPL_POWER_VNOM]/10000;
		}
		else if (dflt->vpp1.present & (1<<CISTPL_POWER_VNOM))
		{
			link->conf.Vpp =
				dflt->vpp1.param[CISTPL_POWER_VNOM]/10000;
		}

		if (cfg->irq.IRQInfo1 || dflt->irq.IRQInfo1)
		{
			 link->conf.Attributes |= CONF_ENABLE_IRQ;
		}

		// IO window settings
		link->io.NumPorts1 = link->io.NumPorts2 = 0;
		if ((cfg->io.nwin > 0) || (dflt->io.nwin > 0))
		{
			cistpl_io_t *io = (cfg->io.nwin) ? &cfg->io : &dflt->io;
			link->io.Attributes1 = IO_DATA_PATH_WIDTH_AUTO;
			if (!(io->flags & CISTPL_IO_8BIT))
			{
				link->io.Attributes1 = IO_DATA_PATH_WIDTH_16;
				DebugLog(KERN_INFO "IO_DATA_PATH_WIDTH_16\n");
			}
			if (!(io->flags & CISTPL_IO_16BIT))
			{
				link->io.Attributes1 = IO_DATA_PATH_WIDTH_8;
				DebugLog(KERN_INFO "IO_DATA_PATH_WIDTH_8\n");
			}
			link->io.IOAddrLines = io->flags & CISTPL_IO_LINES_MASK;
			link->io.BasePort1 = io->win[0].base;
			link->io.NumPorts1 = io->win[0].len;
			if (io->nwin > 1)
			{
				link->io.Attributes2 = link->io.Attributes1;
				link->io.BasePort2 = io->win[1].base;
				link->io.NumPorts2 = io->win[1].len;
			}
			// This reserves IO space but doesn't actually enable it
			if (pcmcia_request_io (link, &link->io) != 0)
			{
				break;
			}
		}
		/*
		Now set up a common memory window, if needed.  There is room
		in the dev_link_t structure for one memory window handle,
		but if the base addresses need to be saved, or if multiple
		windows are needed, the info should go in the private data
		structure for this device.

		Note that the memory window base is a physical address, and
		needs to be mapped to virtual space with ioremap() before it
		is used.
		*/
		if ((cfg->mem.nwin > 0) || (dflt->mem.nwin > 0))
		{
			cistpl_mem_t *mem =
				(cfg->mem.nwin) ? &cfg->mem : &dflt->mem;
			req->Attributes = WIN_DATA_WIDTH_16|WIN_MEMORY_TYPE_CM;
			req->Attributes |= WIN_ENABLE;
			req->Base = mem->win[0].host_addr;
			req->Size = mem->win[0].len;
			if (req->Size < 0x1000)
			{
				req->Size = 0x1000;
			}
			req->AccessSpeed = 0;
			link->win = (window_handle_t) link;
			pcmcia_request_window(&link, req, &link->win);
		}

		nRetVal = 0;

	} while (FALSE);

	return nRetVal;
}

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,27)

struct pcmcia_cfg_mem
{
	tuple_t tuple;
	cisparse_t parse;
	u8 buf[256];
	cistpl_cftable_entry_t dflt;
};

int pcmcia_loop_config (struct pcmcia_device *p_dev, \
	int (*pfnConfigCheck) (struct pcmcia_device *p_dev, \
	cistpl_cftable_entry_t *cfg, cistpl_cftable_entry_t *dflt, \
	unsigned int vcc, void *priv_data), void *priv_data)
{
	struct pcmcia_cfg_mem *cfg_mem;

	tuple_t *tuple;
	int ret = -ENODEV;
	unsigned int vcc;

	cfg_mem = kzalloc (sizeof (struct pcmcia_cfg_mem), GFP_KERNEL);
	if (NULL == cfg_mem)
	{
		return -ENOMEM;
	}

	/* get the current Vcc setting */
	vcc = p_dev->socket->socket.Vcc;

	tuple = &cfg_mem->tuple;
	tuple->TupleData = cfg_mem->buf;
	tuple->TupleDataMax = 255;
	tuple->TupleOffset = 0;
	tuple->DesiredTuple = CISTPL_CFTABLE_ENTRY;
	tuple->Attributes = 0;

	ret = pcmcia_get_first_tuple (p_dev, tuple);
	while (!ret)
	{
		cistpl_cftable_entry_t *cfg = &cfg_mem->parse.cftable_entry;

		if (pcmcia_get_tuple_data (p_dev, tuple))
		{
			goto next_entry;
		}

		if (pcmcia_parse_tuple (p_dev, tuple, &cfg_mem->parse))
		{
			goto next_entry;
		}

		/* default values */
		p_dev->conf.ConfigIndex = cfg->index;
		if (cfg->flags & CISTPL_CFTABLE_DEFAULT)
		{
			cfg_mem->dflt = *cfg;
		}

		ret = pfnConfigCheck (p_dev, cfg, &cfg_mem->dflt, vcc, priv_data);
		if (!ret)
		{
			break;
		}

next_entry:
		ret = pcmcia_get_next_tuple (p_dev, tuple);
	}

	if (cfg_mem != NULL)
	{
		kfree (cfg_mem);
	}

	return ret;
}

#endif

static void Driver_Config( struct pcmcia_device *link )
{
	/*======================================================================
	This function parses through the CIS and prepares the card for IO
	operations.
	======================================================================*/
	tuple_t 	tuple;
	cisparse_t 	parse;
	u_char 		buf[64];
	win_req_t 	req;
	int			nRetVal, i;

	DEVICE_EXTENSION *pDeviceExtension = NULL;

	DebugLog ("Driver_Config : Entering\n");

	for (i=0; i<MAX_PCMCIA_DEVICES; i++)
	{
		if (link == sDeviceExtension[i].DeviceLink)
		{
			pDeviceExtension = &sDeviceExtension [i];
			break;
		}
	}

	if (!pDeviceExtension)
	{
		return;
	}

	DebugLog("\n Driver_Config : Configuring the module \n");

	tuple.DesiredTuple 	= CISTPL_CONFIG;
	tuple.Attributes 	= 0;
	tuple.TupleData 	= buf;
	tuple.TupleDataMax 	= sizeof(buf);
	tuple.TupleOffset 	= 0;

 	CS_CHECK(pcmcia_get_first_tuple, link, &tuple);
 	CS_CHECK(pcmcia_get_tuple_data, link, &tuple);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,28)
	CS_CHECK(pcmcia_parse_tuple, &tuple, &parse);
#else
	CS_CHECK(pcmcia_parse_tuple, link, &tuple, &parse);
#endif

	link->conf.ConfigBase = parse.config.base;
	link->conf.Present = parse.config.rmask[0];

	if (pcmcia_loop_config (link, ConfigCheck, &req) != 0)
	{
		DebugLog("pcmcia_loop_config failed\n");
		goto cs_failed;
	}

	// Allocate an interrupt line.  Note that this does not assign a
	// handler to the interrupt, unless the 'Handler' member of the
	// irq structure is initialized.
	if (link->conf.Attributes & CONF_ENABLE_IRQ)
	{
//		nRetVal = CardServices(RequestIRQ, link->handle, &link->irq);
		nRetVal = pcmcia_request_irq(link, &link->irq);

		if (0 == nRetVal)
		{
			DebugLog(" Interrupt Successfully Assigned \n");
			pDeviceExtension->bIsIrqSet = TRUE;
			pDeviceExtension->bOrgIrqStatus = TRUE;
 			DebugLog (" Driver_Config : ******* IRQ assigned! *******\n");
		}
		else
		{
			pDeviceExtension->bIsIrqSet = FALSE;
			pDeviceExtension->bOrgIrqStatus = FALSE;
		}
		
		if( link->irq.IRQInfo1 & IRQ_INFO2_VALID )
		{
			if( link->irq.IRQInfo2 & IRQ_HANDLE_PRESENT )
			{
				DebugLog(
						" IRQ-HANDLE-PRESENT Macro is set. Handler is being used . \n");
			}
		}
	}

	DebugLog(" INTERRUPT SUCCESSFULLY ASSIGNED \n");
	// This actually configures the PCMCIA socket -- setting up
	// the I/O windows and the interrupt mapping, and putting the
	// card and host interface into "Memory and IO" mode.

	CS_CHECK(pcmcia_request_configuration, link, &link->conf);

	// We can release the IO port allocations here, if some other
	// driver for the card is going to loaded, and will expect the
	// ports to be available.
	if (free_ports)
	{
		if (link->io.BasePort1)
		{
			release_region(link->io.BasePort1, link->io.NumPorts1);
		}
		if (link->io.BasePort2)
		{
			release_region(link->io.BasePort2, link->io.NumPorts2);
		}
	}

	// At this point, the dev_node_t structure(s) need to be
	// initialized and arranged in a linked list at link->dev.
	sprintf(pDeviceExtension->DeviceNode.dev_name, "ACR92%d", pDeviceExtension->nInstanceCount);
	pDeviceExtension->DeviceNode.major = pDeviceExtension->nMajorNumber;	/*ACR92_MAJ_VER;*/
	pDeviceExtension->DeviceNode.minor = pDeviceExtension->nInstanceCount;

	// Maps Physical Memmory To Local Data Structure For IO
	// Using physical Base requires (caddr_t) cast
	pDeviceExtension->am_base = ioremap(req.Base, req.Size);
	pDeviceExtension->IoBase = link->io.BasePort1;

	// Make Device Link
	link->dev_node = &pDeviceExtension->DeviceNode;

	// Finally, report what we've done
	// dev->node.dev_name, link->conf.ConfigIndex,
	// link->conf.Vcc/10, link->conf.Vcc%10);
	if (link->conf.Vpp)
	{
		DebugLog(
				", Vpp %d.%d",
				link->conf.Vpp/10,
				link->conf.Vpp%10
				);
	}
	if (link->conf.Attributes & CONF_ENABLE_IRQ)
	{
		//DebugLog(" irq %d", 
		//		((int)pDeviceExtension->nAssignedIRQ = link->irq.AssignedIRQ));
	}
	if (link->io.NumPorts1)
	{
		DebugLog(", io 0x%04x-0x%04x", link->io.BasePort1,
		link->io.BasePort1+link->io.NumPorts1-1);
	}
	if (link->io.NumPorts2)
	{
		DebugLog(" & 0x%04x-0x%04x", link->io.BasePort2,
		link->io.BasePort2+link->io.NumPorts2-1);
	}
	if (link->win)
	{
		DebugLog(
				", mem 0x%06lx-0x%06lx",
				req.Base,
				req.Base+req.Size-1
				);
	}
	DebugLog("\n");

	Driver_SetIOBase(pDeviceExtension, link->io.BasePort1);

	pDeviceExtension->bSignalEvent = FALSE;

	nRetVal = Driver_InitLocal (pDeviceExtension->nInstanceCount);
	pDeviceExtension->irqFlag = FALSE;
	pDeviceExtension->bIsQueueSet = FALSE;

	if( STATUS_SUCCESS != nRetVal )
	{
		DebugLog(" Device is NOT Present \n");
		return;
	}

	if( FALSE == pDeviceExtension->bIsIrqSet )
	{
		init_waitqueue_head(&pDeviceExtension->tqsCardStatusWait);
		pDeviceExtension->DE_thread_pid = kernel_thread(PcdmCardStatusThread, pDeviceExtension, CLONE_KERNEL);
		DebugLog (" Driver_Config : ********** Thread created **********\n");
		pDeviceExtension->bIsPolledMode = TRUE;

	}

	DebugLog (" Driver_Config EXIT\n");

	return;

	cs_failed:
		DebugLog(" Failure has occured in allocation !!!! \n");
		pcmcia_disable_device(link);

} // ACR92_config

#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)

int Driver_Attach (struct pcmcia_device *p_dev)
{
	dev_link_t	*pDeviceLink;
	int		nCount;

	DEVICE_EXTENSION	*pDeviceExtension = NULL;

	DebugLog ("Inside Driver_Attach ----- \n");

	pDeviceExtension = &sDeviceExtension [nInstanceCount];
	RtlZeroMemory (pDeviceExtension, sizeof(DEVICE_EXTENSION));

//	DebugLog ("Driver_Attach : Sizeof pDeviceExtension : %d", sizeof (DEVICE_EXTENSION));

	pDeviceLink = ( dev_link_t *)kmalloc( sizeof(dev_link_t), GFP_KERNEL );
	if( NULL == pDeviceLink )
	{
	    DebugLog(" Driver Attach : Out of resources!");
	    return -ENOMEM;
	}
	memset ( pDeviceLink , 0, sizeof( dev_link_t ) );

	pDeviceExtension->pucDeviceName = (PUCHAR ) kmalloc( sizeof(UCHAR) * 32, GFP_KERNEL );
	if( NULL == pDeviceExtension->pucDeviceName )
	{
		pDeviceExtension = NULL;
		kfree ( pDeviceLink );
		DebugLog ("Driver_Attach : kmalloc failed for pucDeviceName\n");
		return -ENOMEM;
	}

	pDeviceExtension->DeviceLink = pDeviceLink;

	pDeviceExtension->CardTracker.pDeviceExtension = pDeviceExtension;


	// End of Allocation of memories
	// setting default values to device extension structure.
	pDeviceExtension->nMajorNumber 		= -1;
	pDeviceExtension->nInstanceCount 	= nInstanceCount;
	pDeviceExtension->bBootRomPresent	= FALSE;
	pDeviceExtension->bIsPolledMode 	= FALSE;

	// pcsc daemon does not issue a power request to update
	// card status on system resume from standby/suspend.
	// hence, a card absent/present is simuated from the
	// kernel module.
	pDeviceExtension->bIsSystemResume	= FALSE;

	sprintf (pDeviceExtension->pucDeviceName,"ACR92%d", nInstanceCount);

//	DebugLog (" Driver_Attach of ACR92 Module:: [%s]\n", pDeviceExtension->pucDeviceName);
	pDeviceExtension->nMajorNumber = nMajorNumber;

 	init_waitqueue_head ( &pDeviceExtension->tqsEventWait );
	init_waitqueue_head ( &pDeviceExtension->tqsReadEventWait );

	// Interrupt setup
	pDeviceLink->irq.Attributes = IRQ_TYPE_DYNAMIC_SHARING | IRQ_HANDLE_PRESENT;
	pDeviceLink->irq.IRQInfo1 = IRQ_INFO2_VALID | IRQ_LEVEL_ID;
	if (irq_list[0] == -1)
	{
		pDeviceLink->irq.IRQInfo2 = irq_mask;
	}
	else
	{
		for (nCount = 0; nCount < 4; nCount++)
		{
			pDeviceLink->irq.IRQInfo2 |= 1 << irq_list[nCount];
		}
	}

	pDeviceLink->irq.Handler = &Driver_IrqHandler;
	pDeviceLink->irq.Instance = pDeviceExtension;

	pDeviceLink->conf.Attributes = CONF_ENABLE_IRQ;
	pDeviceLink->conf.Vcc = 50;
	pDeviceLink->conf.IntType = INT_MEMORY_AND_IO;

	// added by Jithu - moving out card insertion event
	// to here from the event handler as there is no
	// event handler from 2.6.16 onwards

	pDeviceLink->handle = p_dev;
	p_dev->instance = pDeviceLink;

	pDeviceLink->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
	Driver_Config(pDeviceLink);

	DebugLog ("Driver_Attach EXIT \n");

	return 0;
}

void Driver_Detach(struct pcmcia_device *p_dev)
{
	int i;	
	dev_link_t *pDeviceLink = dev_to_instance(p_dev);	
	DEVICE_EXTENSION *pDeviceExtension = NULL;

	DebugLog( "\n ACR92::Driver_Detach Entry \n");

	if(!pDeviceLink)
	{
		DebugLog (" NULL pointer for pDeviceLink\n");
		return;
	}

	for (i=0; i<MAX_PCMCIA_DEVICES; i++)
	{
		if ( pDeviceLink == sDeviceExtension[i].DeviceLink )
		{
			pDeviceExtension = &sDeviceExtension[i];
			break;
		}
	}

	if ( !pDeviceExtension )
	{
		DebugLog (" Could not find device extension\n");
	}


	// Card Removal event moved up from event handler to here
	DebugLog("<<<------- card removed -------->>>\n");
	pDeviceLink->state &= ~DEV_PRESENT;

	pDeviceLink->state &= DEV_STALE_CONFIG;

	if (pDeviceLink->state & DEV_CONFIG)
	{
		pDeviceExtension->nStop = 1;
		flush_scheduled_work();
	}

	DebugLog ("\n Driver_Release\n");

	pcmcia_release_configuration(pDeviceLink->handle);
	DebugLog (" Driver_Release : pcmcia_release_configuration done\n");

	// Moved here by Vijay
	if (pDeviceLink->irq.AssignedIRQ)
	{
		// we verify if the IRQ is already released by the DFU
		// operation. if not, only then we release it here
		pcmcia_release_irq(pDeviceLink->handle, &pDeviceLink->irq);
		DebugLog (" Driver_Release : pcmcia_release_irq done\n");
	}
	
	// Don't bother checking to see if these succeed or not
	if (pDeviceLink->win)
	{
		pcmcia_release_window(pDeviceLink->win);
	}

	if (pDeviceLink->io.NumPorts1)
	{
		pcmcia_release_io (pDeviceLink->handle, &pDeviceLink->io);
	}
	
	pDeviceLink->state &= ~DEV_CONFIG;

	DebugLog (" Driver_Detach : Flushing pending work\n");
	flush_scheduled_work ();

	if( NULL != pDeviceExtension )
	{
		pDeviceExtension->bSignalEvent = TRUE;
		pDeviceExtension->bIsEventSet  = TRUE;
		wake_up_interruptible ( &pDeviceExtension->tqsReadEventWait );
	}
	else
	{
		DebugLog (" Driver_Detach : NULL pointer for pDeviceExtension\n");
		return;
	}


	// 	 Kill the card tracking thread
	if( TRUE == pDeviceExtension->bIsPolledMode )
	{
		DebugLog(" ACR92::Stopping the Polled Mode Card Tracking Thread.\n");
		pDeviceExtension->bIsPolledMode = FALSE;
		pDeviceExtension->bIsIrqSet = FALSE;

		send_sig(SIGTERM, pDeviceExtension->pDE_thread, 1);
		pDeviceExtension->pDE_thread = 0;
	}


	if( pDeviceLink->priv )
	{
		kfree(pDeviceLink->priv);
	}

	if( pDeviceExtension->pucDeviceName )
	{
		kfree( pDeviceExtension->pucDeviceName );
		pDeviceExtension->pucDeviceName = NULL;
	}

	if( pDeviceExtension )
	{
		pDeviceExtension = NULL;
	}

	DebugLog(" Driver_Detach Exit\n");
}

int Driver_Suspend(struct pcmcia_device *p_dev)
{
	int i;
	dev_link_t *pDeviceLink = dev_to_instance (p_dev);
	DEVICE_EXTENSION *pDeviceExtension = NULL;

	DebugLog( "\n ACR92::Driver_Suspend Entry \n");


	if(!pDeviceLink)
	{
		DebugLog (" NULL pointer for pDeviceLink\n");
		return -1;
	}

	for (i=0; i<MAX_PCMCIA_DEVICES; i++)
	{
		if ( pDeviceLink == sDeviceExtension[i].DeviceLink )
		{
			pDeviceExtension = &sDeviceExtension[i];
			break;
		}
	}

	if( NULL == pDeviceExtension )
	{
		DebugLog (" Driver_Suspend : NULL pointer for pDeviceExtension\n");
		return -1;
	}

	DebugLog("<<<------- card suspend-------->>>\n");
	pDeviceLink->state |= DEV_SUSPEND;
	// Fall through...
	
		// Mark the device as stopped, to block IO until later
	pDeviceExtension->nStop = 1;
	if (pDeviceLink->state & DEV_CONFIG)
	{
		pcmcia_release_configuration(pDeviceLink->handle);
	}
	
	return 0;
}


int Driver_Resume(struct pcmcia_device *p_dev)
{
	int i;
	dev_link_t *pDeviceLink = dev_to_instance (p_dev);
	DEVICE_EXTENSION *pDeviceExtension = NULL;
	int ncsRetVal=0;

	DebugLog( "\n ACR92::Driver_Resume Entry \n");

	if(!pDeviceLink)
	{
		DebugLog (" NULL pointer for pDeviceLink\n");
		return -1;
	}

	for (i=0; i<MAX_PCMCIA_DEVICES; i++)
	{
		if ( pDeviceLink == sDeviceExtension[i].DeviceLink )
		{
			pDeviceExtension = &sDeviceExtension[i];
			break;
		}
	}

	if( NULL == pDeviceExtension )
	{
		DebugLog (" Driver_Resume : NULL pointer for pDeviceExtension\n");
		return -1;
	}

	DebugLog("<<<------- card resume-------->>>\n");
	DebugLog(" Event PM_RESUME Called \n");
	pDeviceLink->state &= ~DEV_SUSPEND;

	//
	//If the Irq was not assigned or if polled mode has been set
	//then start the thread explicitly on resume
	//
	if(FALSE == pDeviceExtension->bIsIrqSet  ||TRUE == pDeviceExtension->bIsPolledMode)
	{
		DebugLog("******* THREAD TO BE STARTED ON RESUME*****\n");
		init_waitqueue_head(&pDeviceExtension->tqsCardStatusWait);
		pDeviceExtension->DE_thread_pid = kernel_thread(PcdmCardStatusThread, pDeviceExtension, CLONE_KERNEL);
	}
	
	DebugLog("<<<------- card reset-------->>>\n");
	if (pDeviceLink->state & DEV_CONFIG)
	{
		pcmcia_request_configuration(pDeviceLink->handle, &pDeviceLink->conf);
	}

	// since the device went to standby/suspend, the firmware
	// will get reset to Rev A status. Hence, we need to update
	// the transport mechanism used prior to suspend/standby to the firmware.


		if( PCDM_REV_C == pDeviceExtension->byPcdmRev )
		{
			DebugLog("*********** Found REVC ON RESUME***************\n");
			// we set the transport mechanism to REv-A since the mode will 
			// now be A and kernel module shud also be called by A only.
			pDeviceExtension->byPcdmRev = PCDM_REV_A;
			ncsRetVal = IFDSetTransportMechanism( pDeviceExtension );
			if( 0 == ncsRetVal )
			{
				DebugLog (" Setting PCDM revision to REV C\n");
				pDeviceExtension->byPcdmRev = PCDM_REV_C;
			}
			else
			{
				// we failed to the tranport mechanism. hence, we revert
				// back to rev-A. we dont take into account the return
				// state
				DebugLog (" Setting PCDM revision to REV A\n");
				pDeviceExtension->byPcdmRev = PCDM_REV_A;
				ncsRetVal = IFDSetTransportMechanism( pDeviceExtension );
				DebugLog("Failure in Setting transport Mechanism to Rev-C. Revert to Rev-A\n");
				//break;
			}
		}

		DebugLog (" Event Handler [CS_EVENT_CARD_RESET] : IFDGetCardStatus\n");
		IFDGetCardStatus(pDeviceExtension);
		pDeviceExtension->bIsSystemResume = TRUE;
	
	//In a normal driver, additional code may go here to restore
	//the device state and restart IO.
	pDeviceExtension->nStop = 0;
	return 0;
}

#else

dev_link_t *Driver_Attach( void )
{
	dev_link_t	*pDeviceLink;
	client_reg_t	client_reg;
	int		ncsRetVal;
	int		nCount;

	DEVICE_EXTENSION	*pDeviceExtension = NULL;

	DebugLog ("Inside Driver_Attach ----- \n");

	pDeviceExtension = &sDeviceExtension [nInstanceCount];
	RtlZeroMemory (pDeviceExtension, sizeof(DEVICE_EXTENSION));

//	DebugLog ("Driver_Attach : Sizeof pDeviceExtension : %d", sizeof (DEVICE_EXTENSION));

	pDeviceLink = ( dev_link_t *)kmalloc( sizeof(dev_link_t), GFP_KERNEL );
	if( NULL == pDeviceLink )
	{
	    DebugLog(" Driver Attach : Out of resources!");
	    return NULL;
	}
	memset ( pDeviceLink , 0, sizeof( dev_link_t ) );

	pDeviceExtension->pucDeviceName = (PUCHAR ) kmalloc( sizeof(UCHAR) * 32, GFP_KERNEL );
	if( NULL == pDeviceExtension->pucDeviceName )
	{
		pDeviceExtension = NULL;
		kfree ( pDeviceLink );
		DebugLog ("Driver_Attach : kmalloc failed for pucDeviceName\n");
		return NULL;
	}

	pDeviceExtension->DeviceLink = pDeviceLink;

	pDeviceExtension->CardTracker.pDeviceExtension = pDeviceExtension;


	// End of Allocation of memories
	// setting default values to device extension structure.
	pDeviceExtension->nMajorNumber 		= -1;
	pDeviceExtension->nInstanceCount 	= nInstanceCount;
	pDeviceExtension->bBootRomPresent	= FALSE;
	pDeviceExtension->bIsPolledMode 	= FALSE;

	// pcsc daemon does not issue a power request to update
	// card status on system resume from standby/suspend.
	// hence, a card absent/present is simuated from the
	// kernel module.
	pDeviceExtension->bIsSystemResume	= FALSE;

	sprintf (pDeviceExtension->pucDeviceName,"ACR92%d", nInstanceCount);

//	DebugLog (" Driver_Attach of ACR92 Module:: [%s]\n", pDeviceExtension->pucDeviceName);
	pDeviceExtension->nMajorNumber = nMajorNumber;

 	init_waitqueue_head ( &pDeviceExtension->tqsEventWait );
	init_waitqueue_head ( &pDeviceExtension->tqsReadEventWait );

	// Interrupt setup
	pDeviceLink->irq.Attributes = IRQ_TYPE_DYNAMIC_SHARING | IRQ_HANDLE_PRESENT;
	pDeviceLink->irq.IRQInfo1 = IRQ_INFO2_VALID | IRQ_LEVEL_ID;
	if (irq_list[0] == -1)
	{
		pDeviceLink->irq.IRQInfo2 = irq_mask;
	}
	else
	{
		for (nCount = 0; nCount < 4; nCount++)
		{
			pDeviceLink->irq.IRQInfo2 |= 1 << irq_list[nCount];
		}
	}

	pDeviceLink->irq.Handler = &Driver_IrqHandler;
	pDeviceLink->irq.Instance = pDeviceExtension;

	pDeviceLink->conf.Attributes = CONF_ENABLE_IRQ;
	pDeviceLink->conf.Vcc = 50;
	pDeviceLink->conf.IntType = INT_MEMORY_AND_IO;

	// 	Register with Card Services
	pDeviceLink->next = dev_list;
	dev_list = pDeviceLink;

	client_reg.dev_info = &dev_info;
	client_reg.Attributes = INFO_IO_CLIENT | INFO_CARD_SHARE;
	client_reg.EventMask = 	CS_EVENT_CARD_INSERTION | CS_EVENT_CARD_REMOVAL |
							CS_EVENT_RESET_PHYSICAL | CS_EVENT_CARD_RESET |
							CS_EVENT_PM_SUSPEND | CS_EVENT_PM_RESUME;
	client_reg.event_handler = &Event_Handler;
	client_reg.Version = 0x0100;
	client_reg.event_callback_args.client_data = pDeviceLink;

	ncsRetVal = pcmcia_register_client(&pDeviceLink->handle, &client_reg);

	if (ncsRetVal != CS_SUCCESS)
	{
		DebugLog (" Driver_Attach ACR92 Module FAILED!!!:: (%d) [%s]\n", ncsRetVal,  pDeviceExtension->pucDeviceName);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11)
		cs_error(acr92_ids, RegisterClient, ncsRetVal);
#else
		cs_error(pDeviceLink->handle, RegisterClient, ncsRetVal);
#endif
		Driver_Detach (pDeviceLink);
		return NULL;
	}

// 	nInstanceCount++;

	DebugLog ("Driver_Attach EXIT \n");

	return pDeviceLink;
}

void Driver_Detach(dev_link_t *pDeviceLink)
{
	int i;
	dev_link_t **dpDeviceLink;
	DEVICE_EXTENSION *pDeviceExtension = NULL;

	DebugLog( "\n ACR92::Driver_Detach Entry \n");

	if(!pDeviceLink)
	{
		DebugLog (" NULL pointer for pDeviceLink\n");
		return;
	}

	for (i=0; i<MAX_PCMCIA_DEVICES; i++)
	{
		if ( pDeviceLink == sDeviceExtension[i].DeviceLink )
		{
			pDeviceExtension = &sDeviceExtension[i];
			break;
		}
	}

	if ( !pDeviceExtension )
	{
		DebugLog (" Could not find device extension\n");
	}

	DebugLog (" Driver_Detach : Flushing pending work\n");
	flush_scheduled_work ();

	if( NULL != pDeviceExtension )
	{
		pDeviceExtension->bSignalEvent = TRUE;
		pDeviceExtension->bIsEventSet  = TRUE;
		wake_up_interruptible ( &pDeviceExtension->tqsReadEventWait );
	}
	else
	{
		DebugLog (" Driver_Detach : NULL pointer for pDeviceExtension\n");
		return;
	}

	// 	 Locate device structure
	for (dpDeviceLink = &dev_list; *dpDeviceLink; dpDeviceLink = &(*dpDeviceLink)->next)
	{
		if ( *dpDeviceLink == pDeviceLink )
		{
			break;
		}
	}
	if (*dpDeviceLink == NULL)
	{
		return;
	}

	if (pDeviceLink->state & DEV_CONFIG)
	{
		pDeviceLink->state |= DEV_STALE_LINK;
		return;
	}

	// 	 Kill the card tracking thread
	if( TRUE == pDeviceExtension->bIsPolledMode )
	{
		DebugLog(" ACR92::Stopping the Polled Mode Card Tracking Thread.\n");
		pDeviceExtension->bIsPolledMode = FALSE;
		pDeviceExtension->bIsIrqSet = FALSE;

		send_sig(SIGTERM, pDeviceExtension->pDE_thread, 1);
		pDeviceExtension->pDE_thread = 0;
	}

	// 	 Break the link with Card Services
	if (pDeviceLink->handle)
	{
		pcmcia_deregister_client(pDeviceLink->handle);
		DebugLog (" Driver_Detach : Deregistered Client!\n");
	}

	// 	 Unlink device structure, and free it
	*dpDeviceLink = pDeviceLink->next;

	if( pDeviceLink->priv )
	{
		kfree(pDeviceLink->priv);
	}

	if( pDeviceExtension->pucDeviceName )
	{
		kfree( pDeviceExtension->pucDeviceName );
		pDeviceExtension->pucDeviceName = NULL;
	}

	if( pDeviceExtension )
	{
		pDeviceExtension = NULL;
	}

	DebugLog(" Driver_Detach Exit\n");
}
void Driver_Release(u_long arg)
{
	// After a card is removed, ACR92_release() will unregister the
	// device, and release the PCMCIA configuration.  If the device is
	// still open, this will be postponed until it is closed.
	dev_link_t *pDeviceLink = (dev_link_t *)arg;

	DebugLog ("\n Driver_Release\n");

	pcmcia_release_configuration(pDeviceLink->handle);
	DebugLog (" Driver_Release : pcmcia_release_configuration done\n");

	// Moved here by Vijay
	if (pDeviceLink->irq.AssignedIRQ)
	{
		// we verify if the IRQ is already released by the DFU
		// operation. if not, only then we release it here
		pcmcia_release_irq(pDeviceLink->handle, &pDeviceLink->irq);
		DebugLog (" Driver_Release : pcmcia_release_irq done\n");
	}

	// If the device is currently in use, we won't release until it
	// is actually closed, because until then, we can't be sure that
	// no one will try to access the device or its data structures.
	if (pDeviceLink->open)
	{
		pDeviceLink->state |= DEV_STALE_CONFIG;
		return;
	}

	// Unlink the device chain
// 	pDeviceLink->dev = NULL;

	// In a normal driver, additional code may be needed to release
	// other kernel data structures associated with this device.

	// Don't bother checking to see if these succeed or not
	if (pDeviceLink->win)
	{
		pcmcia_release_window(pDeviceLink->win);
	}

// 	pcmcia_release_configuration(pDeviceLink->handle);

	if (pDeviceLink->io.NumPorts1)
	{
		pcmcia_release_io (pDeviceLink->handle, &pDeviceLink->io);
	}
	/*
	if (pDeviceLink->irq.AssignedIRQ)
	{
		// we verify if the IRQ is already released by the DFU
		// operation. if not, only then we release it here
		pcmcia_release_irq(pDeviceLink->handle, &pDeviceLink->irq);
	}*/
	pDeviceLink->state &= ~DEV_CONFIG;

	if (pDeviceLink->state & DEV_STALE_LINK)
	{
		DebugLog("Detaching from Driver_Release \n");
		Driver_Detach(pDeviceLink);
	}

	// Unlink the device chain
 	pDeviceLink->dev = NULL;

	DebugLog("Driver_Release Exit\n");
}

int Event_Handler(
	event_t event, int priority,
	event_callback_args_t *args
	)
{
	/*======================================================================
	The card status event handler.  Mostly, this schedules other
	stuff to run after an event is received.

	When a CARD_REMOVAL event is received, we immediately set a
	private flag to block future accesses to this device.  All the
	functions that actually access the device should check this flag
	to make sure the card is still present.
	======================================================================*/

	DEVICE_EXTENSION *pDeviceExtension = NULL;

	dev_link_t *pDeviceLink = args->client_data;
	int i, ncsRetVal=0;

	//DebugLog ("\n Event_Handler Entry with pDeviceLink %ld\n", pDeviceLink);

	for (i=0; i<MAX_PCMCIA_DEVICES; i++)
	{
		if ( pDeviceLink == sDeviceExtension[i].DeviceLink )
		{
			pDeviceExtension = &sDeviceExtension[i];
			break;
		}
	}

	if (NULL == pDeviceExtension)
	{
		return -1;
	}

	switch (event) {
	case CS_EVENT_CARD_REMOVAL:
		DebugLog("<<<------- card removed -------->>>\n");
		pDeviceLink->state &= ~DEV_PRESENT;

		pDeviceLink->state &= DEV_STALE_CONFIG;

		if (pDeviceLink->state & DEV_CONFIG)
		{
			pDeviceExtension->nStop = 1;
			flush_scheduled_work();
		}

		Driver_Release ((u_long) pDeviceLink);

		break;

	case CS_EVENT_CARD_INSERTION:
		DebugLog("<<<------- card inserted-------->>>\n");
		pDeviceLink->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
		Driver_Config(pDeviceLink);
		break;

	case CS_EVENT_PM_SUSPEND:
		DebugLog("<<<------- card suspend-------->>>\n");
		pDeviceLink->state |= DEV_SUSPEND;
		// Fall through...
	case CS_EVENT_RESET_PHYSICAL:
		// Mark the device as stopped, to block IO until later
		pDeviceExtension->nStop = 1;
		if (pDeviceLink->state & DEV_CONFIG)
		{
			pcmcia_release_configuration(pDeviceLink->handle);
		}
		break;

	case CS_EVENT_PM_RESUME:
		DebugLog("<<<------- card resume-------->>>\n");
		DebugLog(" Event PM_RESUME Called \n");
		pDeviceLink->state &= ~DEV_SUSPEND;

		//
		//If the Irq was not assigned or if polled mode has been set
		//then start the thread explicitly on resume
		//
		if(FALSE == pDeviceExtension->bIsIrqSet  ||TRUE == pDeviceExtension->bIsPolledMode)
		{
			DebugLog("******* THREAD TO BE STARTED ON RESUME*****\n");
			init_waitqueue_head(&pDeviceExtension->tqsCardStatusWait);
			pDeviceExtension->DE_thread_pid = kernel_thread(PcdmCardStatusThread, pDeviceExtension, CLONE_KERNEL);
		}
		/* Fall through... */
	case CS_EVENT_CARD_RESET:
		DebugLog("<<<------- card reset-------->>>\n");
		if (pDeviceLink->state & DEV_CONFIG)
		{
			pcmcia_request_configuration(pDeviceLink->handle, &pDeviceLink->conf);
		}

		// since the device went to standby/suspend, the firmware
		// will get reset to Rev A status. Hence, we need to update
		// the transport mechanism used prior to suspend/standby to the firmware.
		if( CS_EVENT_PM_RESUME == event )
		{
			if( PCDM_REV_C == pDeviceExtension->byPcdmRev )
			{
				DebugLog("*********** Found REVC ON RESUME***************\n");
				// we set the transport mechanism to REv-A since the mode will 
				// now be A and kernel module shud also be called by A only.
				pDeviceExtension->byPcdmRev = PCDM_REV_A;
				ncsRetVal = IFDSetTransportMechanism( pDeviceExtension );
				if( 0 == ncsRetVal )
				{
					DebugLog (" Setting PCDM revision to REV C\n");
					pDeviceExtension->byPcdmRev = PCDM_REV_C;
				}
				else
				{
					// we failed to the tranport mechanism. hence, we revert
					// back to rev-A. we dont take into account the return
					// state
					DebugLog (" Setting PCDM revision to REV A\n");
					pDeviceExtension->byPcdmRev = PCDM_REV_A;
					ncsRetVal = IFDSetTransportMechanism( pDeviceExtension );
					DebugLog("Failure in Setting transport Mechanism to Rev-C. Revert to Rev-A\n");
					break;
				}
			}

			DebugLog (" Event Handler [CS_EVENT_CARD_RESET] : IFDGetCardStatus\n");
			IFDGetCardStatus(pDeviceExtension);
			pDeviceExtension->bIsSystemResume = TRUE;
		}
		//In a normal driver, additional code may go here to restore
		//the device state and restart IO.
		pDeviceExtension->nStop = 0;
		break;
	}

	DebugLog (" Event_Handler EXIT\n");

	return ncsRetVal;
} /* ACR92_event */

#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17)

static void Driver_Config( dev_link_t *link )
{
	/*======================================================================
	This function parses through the CIS and prepares the card for IO
	operations.
	======================================================================*/
	client_handle_t handle 	= link->handle;
	tuple_t 	tuple;
	cisparse_t 	parse;
//	int 		last_fn,
//			last_ret;
	u_char 		buf[64];
	config_info_t 	conf;
	win_req_t 	req;
	memreq_t 	map;
	cisinfo_t 	cisinfo;
	int 		ret=0;
	int			nRetVal, i;

	DEVICE_EXTENSION *pDeviceExtension = NULL;

	DebugLog ("Driver_Config : Entering\n");

	for (i=0; i<MAX_PCMCIA_DEVICES; i++)
	{
		if (link == sDeviceExtension[i].DeviceLink)
		{
			pDeviceExtension = &sDeviceExtension [i];
			break;
		}
	}

	if (!pDeviceExtension)
	{
		return;
	}

	DebugLog("\n Driver_Config : Configuring the module \n");

	// This reads the card's CONFIG tuple to find its
	// configuration registers.
	ret = pcmcia_validate_cis(handle,&cisinfo);

	tuple.DesiredTuple 	= CISTPL_CONFIG;
	tuple.Attributes 	= 0;
	tuple.TupleData 	= buf;
	tuple.TupleDataMax 	= sizeof(buf);
	tuple.TupleOffset 	= 0;

 	CS_CHECK(pcmcia_get_first_tuple, handle, &tuple);
 	CS_CHECK(pcmcia_get_tuple_data, handle, &tuple);
	CS_CHECK(pcmcia_parse_tuple, handle, &tuple, &parse);

	link->conf.ConfigBase = parse.config.base;
	link->conf.Present = parse.config.rmask[0];

	// Configure card
	link->state |= DEV_CONFIG;

	// Look up the current Vcc
	CS_CHECK(pcmcia_get_configuration_info, handle, &conf);

	link->conf.Vcc = conf.Vcc;

	/*
	In this loop, we scan the CIS for configuration table entries,
	each of which describes a valid card configuration, including
	voltage, IO window, memory window, and interrupt settings.

	We make no assumptions about the card to be configured: we use
	just the information available in the CIS.  In an ideal world,
	this would work for any PCMCIA card, but it requires a complete
	and accurate CIS.  In practice, a driver usually "knows" most of
	these things without consulting the CIS, and most client drivers
	will only use the CIS to fill in implementation-defined details.
	*/
	tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
	CS_CHECK(pcmcia_get_first_tuple, handle, &tuple);

	while (1)
	{
		cistpl_cftable_entry_t dflt = { 0 };
		cistpl_cftable_entry_t *cfg = &(parse.cftable_entry);

		CFG_CHECK(pcmcia_get_tuple_data, handle, &tuple);
		CFG_CHECK(pcmcia_parse_tuple, handle, &tuple, &parse);

		if (cfg->flags & CISTPL_CFTABLE_DEFAULT)
		{
			dflt = *cfg;
		}
		if (cfg->index == 0)
		{
			goto next_entry;
		}
		link->conf.ConfigIndex = cfg->index;

		// Does this card need audio output?
		if (cfg->flags & CISTPL_CFTABLE_AUDIO)
		{
			link->conf.Attributes |= CONF_ENABLE_SPKR;
			link->conf.Status = CCSR_AUDIO_ENA;
		}

		// Use power settings for Vcc and Vpp if present
		// Note that the CIS values need to be rescaled
		if (cfg->vcc.present & (1<<CISTPL_POWER_VNOM))
		{
			if (conf.Vcc != cfg->vcc.param[CISTPL_POWER_VNOM]/10000)
			{
				goto next_entry;
			}
		}
		else if (dflt.vcc.present & (1<<CISTPL_POWER_VNOM))
		{
			if (conf.Vcc != dflt.vcc.param[CISTPL_POWER_VNOM]/10000)
			{
				goto next_entry;
			}
		}

		if (cfg->vpp1.present & (1<<CISTPL_POWER_VNOM))
		{
			link->conf.Vpp1 = link->conf.Vpp2 =
				cfg->vpp1.param[CISTPL_POWER_VNOM]/10000;
		}
		else if (dflt.vpp1.present & (1<<CISTPL_POWER_VNOM))
		{
			link->conf.Vpp1 = link->conf.Vpp2 =
				dflt.vpp1.param[CISTPL_POWER_VNOM]/10000;
		}

		if (cfg->irq.IRQInfo1 || dflt.irq.IRQInfo1)
		{
			 link->conf.Attributes |= CONF_ENABLE_IRQ;
		}

		// IO window settings
		link->io.NumPorts1 = link->io.NumPorts2 = 0;
		if ((cfg->io.nwin > 0) || (dflt.io.nwin > 0))
		{
			cistpl_io_t *io = (cfg->io.nwin) ? &cfg->io : &dflt.io;
			link->io.Attributes1 = IO_DATA_PATH_WIDTH_AUTO;
			if (!(io->flags & CISTPL_IO_8BIT))
			{
				link->io.Attributes1 = IO_DATA_PATH_WIDTH_16;
				DebugLog(KERN_INFO "IO_DATA_PATH_WIDTH_16\n");
			}
			if (!(io->flags & CISTPL_IO_16BIT))
			{
				link->io.Attributes1 = IO_DATA_PATH_WIDTH_8;
				DebugLog(KERN_INFO "IO_DATA_PATH_WIDTH_8\n");
			}
			link->io.IOAddrLines = io->flags & CISTPL_IO_LINES_MASK;
			link->io.BasePort1 = io->win[0].base;
			link->io.NumPorts1 = io->win[0].len;
			if (io->nwin > 1)
			{
				link->io.Attributes2 = link->io.Attributes1;
				link->io.BasePort2 = io->win[1].base;
				link->io.NumPorts2 = io->win[1].len;
			}
			// This reserves IO space but doesn't actually enable it
			CFG_CHECK(pcmcia_request_io, link->handle, &link->io);
		}
		/*
		Now set up a common memory window, if needed.  There is room
		in the dev_link_t structure for one memory window handle,
		but if the base addresses need to be saved, or if multiple
		windows are needed, the info should go in the private data
		structure for this device.

		Note that the memory window base is a physical address, and
		needs to be mapped to virtual space with ioremap() before it
		is used.
		*/
		if ((cfg->mem.nwin > 0) || (dflt.mem.nwin > 0))
		{
			cistpl_mem_t *mem =
				(cfg->mem.nwin) ? &cfg->mem : &dflt.mem;
			req.Attributes = WIN_DATA_WIDTH_16|WIN_MEMORY_TYPE_CM;
			req.Attributes |= WIN_ENABLE;
			req.Base = mem->win[0].host_addr;
			req.Size = mem->win[0].len;
			if (req.Size < 0x1000)
			{
				req.Size = 0x1000;
			}
			req.AccessSpeed = 0;
			link->win = (window_handle_t)link->handle;
//			CFG_CHECK(RequestWindow, &link->win, &req);
			pcmcia_request_window(&handle, &req, &link->win);

			map.Page = 0; map.CardOffset = mem->win[0].card_addr;
		}

		break;

	next_entry:
		if (link->io.NumPorts1)
		{
			pcmcia_release_io(link->handle, &link->io);
		}

		pcmcia_get_next_tuple(handle, &tuple);
	}

	// Allocate an interrupt line.  Note that this does not assign a
	// handler to the interrupt, unless the 'Handler' member of the
	// irq structure is initialized.
	if (link->conf.Attributes & CONF_ENABLE_IRQ)
	{
//		nRetVal = CardServices(RequestIRQ, link->handle, &link->irq);
		nRetVal = pcmcia_request_irq(link->handle, &link->irq);

		if( nRetVal == CS_SUCCESS )
		{
			DebugLog(" Interrupt Successfully Assigned \n");
			pDeviceExtension->bIsIrqSet = TRUE;
			pDeviceExtension->bOrgIrqStatus = TRUE;
 			DebugLog (" Driver_Config : ******* IRQ assigned! *******\n");
		}
		else
		{
			pDeviceExtension->bIsIrqSet = FALSE;
			pDeviceExtension->bOrgIrqStatus = FALSE;
		}
		if( nRetVal & CS_IN_USE )
		{
			DebugLog(" Requested Interupt not available, \
					or already assigned \n");
		}
		if( nRetVal & CS_BAD_ATTRIBUTE )
		{
			DebugLog(" BAD ATTRIBUTE SET \n");
		}
		if( nRetVal & CS_CONFIGURATION_LOCKED )
		{
			DebugLog(" Driver_SetIrq Entry !!! \n");
			DebugLog(" CONFIGURATION IS ALREADY LOCKED \n");
		}

		if( link->irq.IRQInfo1 & IRQ_INFO2_VALID )
		{
			if( link->irq.IRQInfo2 & IRQ_HANDLE_PRESENT )
			{
				DebugLog(
						" IRQ-HANDLE-PRESENT Macro is set. Handler is being used . \n");
			}
		}
	}

	DebugLog(" INTERRUPT SUCCESSFULLY ASSIGNED \n");
	// This actually configures the PCMCIA socket -- setting up
	// the I/O windows and the interrupt mapping, and putting the
	// card and host interface into "Memory and IO" mode.

	CS_CHECK(pcmcia_request_configuration, link->handle, &link->conf);

	// We can release the IO port allocations here, if some other
	// driver for the card is going to loaded, and will expect the
	// ports to be available.
	if (free_ports)
	{
		if (link->io.BasePort1)
		{
			release_region(link->io.BasePort1, link->io.NumPorts1);
		}
		if (link->io.BasePort2)
		{
			release_region(link->io.BasePort2, link->io.NumPorts2);
		}
	}

	// At this point, the dev_node_t structure(s) need to be
	// initialized and arranged in a linked list at link->dev.
	sprintf(pDeviceExtension->DeviceNode.dev_name, "ACR92%d", pDeviceExtension->nInstanceCount);
	pDeviceExtension->DeviceNode.major = pDeviceExtension->nMajorNumber;	/*ACR92_MAJ_VER;*/
	pDeviceExtension->DeviceNode.minor = pDeviceExtension->nInstanceCount;

	// Maps Physical Memmory To Local Data Structure For IO
	// Using physical Base requires (caddr_t) cast
	pDeviceExtension->am_base = ioremap(req.Base, req.Size);
	pDeviceExtension->IoBase = link->io.BasePort1;

	// Make Device Link
	link->dev = &pDeviceExtension->DeviceNode;

	// Finally, report what we've done
	// dev->node.dev_name, link->conf.ConfigIndex,
	// link->conf.Vcc/10, link->conf.Vcc%10);
	if (link->conf.Vpp1)
	{
		DebugLog(
				", Vpp %d.%d",
				link->conf.Vpp1/10,
				link->conf.Vpp1%10
				);
	}
	if (link->conf.Attributes & CONF_ENABLE_IRQ)
	{
		//DebugLog(" irq %d", 
		//		((int)pDeviceExtension->nAssignedIRQ = link->irq.AssignedIRQ));
	}
	if (link->io.NumPorts1)
	{
		DebugLog(", io 0x%04x-0x%04x", link->io.BasePort1,
		link->io.BasePort1+link->io.NumPorts1-1);
	}
	if (link->io.NumPorts2)
	{
		DebugLog(" & 0x%04x-0x%04x", link->io.BasePort2,
		link->io.BasePort2+link->io.NumPorts2-1);
	}
	if (link->win)
	{
		DebugLog(
				", mem 0x%06lx-0x%06lx",
				req.Base,
				req.Base+req.Size-1
				);
	}
	DebugLog("\n");
	link->state &= ~DEV_CONFIG_PENDING;

	Driver_SetIOBase(pDeviceExtension, link->io.BasePort1);

	pDeviceExtension->bSignalEvent = FALSE;

	nRetVal = Driver_InitLocal (pDeviceExtension->nInstanceCount);
	pDeviceExtension->irqFlag = FALSE;
	pDeviceExtension->bIsQueueSet = FALSE;

	if( STATUS_SUCCESS != nRetVal )
	{
		DebugLog(" Device is NOT Present \n");
		return;
	}

	if( FALSE == pDeviceExtension->bIsIrqSet )
	{
		init_waitqueue_head(&pDeviceExtension->tqsCardStatusWait);
		pDeviceExtension->DE_thread_pid = kernel_thread(PcdmCardStatusThread, pDeviceExtension, CLONE_KERNEL);
		DebugLog (" Driver_Config : ********** Thread created **********\n");
		pDeviceExtension->bIsPolledMode = TRUE;

	}

	DebugLog (" Driver_Config EXIT\n");

	return;

	cs_failed:
		DebugLog(" Failure has occured in allocation !!!! \n");

		// Since Driver_Release function is not compiled with
		// future versions, releasing the configurations
		// straight away
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
		Driver_Release((u_long)link);
#else
		pcmcia_release_configuration(link->handle);
		DebugLog (" Equalent to driver release\n");

	
		if (link->irq.AssignedIRQ)
		{	
			pcmcia_release_irq(link->handle, &link->irq);	
		}

		if (link->win)
		{
			pcmcia_release_window(link->win);
		}

		if (link->io.NumPorts1)
		{
			pcmcia_release_io (link->handle, &link->io);
		}
	
		link->state &= ~DEV_CONFIG;

#endif


} // ACR92_config

#endif

void Driver_SetIOBase(DEVICE_EXTENSION *pDeviceExtension,u_long iobase)
{
	//====================================================================
	//	copy the iobase to the device extension
	//====================================================================

	//DebugLog ("Setting IoBase [%ul] for pDeviceExtension [%ld]\n", iobase, pDeviceExtension);

	pDeviceExtension->IoBase = iobase;

}

NTSTATUS Driver_InitLocal(int nCurrentInstanceCount)
{
	NTSTATUS NTStatus = STATUS_SUCCESS;
	DEVICE_EXTENSION	*pDeviceExtension = &sDeviceExtension[nCurrentInstanceCount];

	//====================================================================
	//		Initialization of the device extension struction
	//====================================================================
	DebugLog ("\n ACR92:: Driver_InitLocal Entry of the Kernel Module\n");

	do
	{
		pDeviceExtension->irqFlag		= 0;
		pDeviceExtension->inIOCTL 		= 0;
		pDeviceExtension->prevCardStatus 	= 0;
		pDeviceExtension->cardStsChange		= FALSE;
		pDeviceExtension->cmd_lock 		= SPIN_LOCK_UNLOCKED;
		pDeviceExtension->ioctl_lock		= SPIN_LOCK_UNLOCKED;
		pDeviceExtension->rdwr_lock		= SPIN_LOCK_UNLOCKED;

		pDeviceExtension->byPcdmRev 		= PCDM_REV_A;

		spin_lock_init(&pDeviceExtension->cmd_lock);
		spin_lock_init(&pDeviceExtension->ioctl_lock);
		spin_lock_init(&pDeviceExtension->rdwr_lock);

		init_waitqueue_head(&pDeviceExtension->wait);
		init_waitqueue_head(&pDeviceExtension->delay);

		NTStatus = IFDGetInterfaceDetails( pDeviceExtension );
		if( STATUS_SUCCESS != NTStatus )
		{
			pDeviceExtension->bBootRomPresent = TRUE;
			DebugLog(" the device is in BOOT ROM MODE \n");
			break;
		}

		DebugLog (" Driver_InitLocal : IFDGetCardStatus\n");
		NTStatus = IFDGetCardStatus(pDeviceExtension);
	}while(FALSE);

	DebugLog (" ACR92:: Driver_InitLocal Exit of the Kernel Module [%ld]\n", NTStatus);

	return NTStatus;
}

void CardStatusCheck(void  *arg)
{
	//====================================================================
	//		Checks the presence of the smart card
	//====================================================================
	DEVICE_EXTENSION	*pDeviceExtension =  *(DEVICE_EXTENSION **)arg;

	DebugLog ("\nACR92:: CardStatusCheck Entry of Kernel Module\n");
	spin_lock(&pDeviceExtension->cmd_lock);
	pDeviceExtension->irqFlag = FALSE;

	DebugLog (" CardStatusCheck : IFDGetCardStatus\n");
	IFDGetCardStatus(pDeviceExtension);
	spin_unlock(&pDeviceExtension->cmd_lock);

	DebugLog ("ACR92:: CardStatusCheck EXIT of Kernel Module\n");
}



int IFDHOpen(struct inode *inode , struct file *file )
{
	//======================================================================
	// 	This function is called when the IfdHandler opens the device.
	//=======================================================================

	unsigned int minor = iminor(inode);
	DEVICE_EXTENSION *pDeviceExtension = NULL;

	DebugLog ("\n Calling IFDHOpen of the KERNEL MODULE OF ACR92 for minor (%d)\n", minor);

	pDeviceExtension = &sDeviceExtension [minor];

	//if (ps && ps->pcmcia)
	if (pDeviceExtension)
	{
		try_module_get (THIS_MODULE);
 		DebugLog (" IFDHOpen : Module reference count [%u]\n", module_refcount(THIS_MODULE));
		minor = 0;
	}
	else
	{
		minor = -ENODEV;
	}

	DebugLog (" IFDHOpen EXIT [%d]\n", minor);

	return minor;
}


int IFDHClose(struct inode *inode , struct file *file )
{
	//======================================================================
	//	Called when the Ifdhandler closes communication with the device
	//======================================================================

	DEVICE_EXTENSION	*pDeviceExtension = NULL;

	DebugLog ("\n Calling IFDHClose of the KERNEL MODULE OF ACR92\n");

	pDeviceExtension = &sDeviceExtension [ iminor(inode) ];

	do
	{
		if( NULL == pDeviceExtension )
		{
			break;
		}

		if ( !file )
		{
		    DebugLog (" Invalid File Handle!\n");
		}

		if( TRUE == pDeviceExtension->bIsPolledMode )
		{
			DebugLog(" Killing Polled Mode Card Tracking from IFDHClose %d\n", pDeviceExtension->bOrgIrqStatus);

			pDeviceExtension->bIsPolledMode = FALSE;
			pDeviceExtension->bIsIrqSet = pDeviceExtension->bOrgIrqStatus;

			if( TRUE == pDeviceExtension->bOrgIrqStatus )
			{
				/* Kill the card tracking thread */
				DebugLog(" STOPPING CARD TRACKING THREAD AS IRQ is absent \n" );
				send_sig(SIGTERM, pDeviceExtension->pDE_thread, 1);
			}
		}
		//	Added by Vijay. Decrements Module usage count.
		module_put (THIS_MODULE);
		DebugLog (" Reference count is now : %u\n", module_refcount(THIS_MODULE));

	}while( FALSE );

	DebugLog (" IFDHClose EXIT\n");

	return 0;
}


int IFDHIoctl(struct inode *inode, struct file *file, u_int ucIoctl, u_long arg)
{
	//======================================================================
	//				ioctl calls from the ifdhandler.
	//				cmd is the IOCTL code
	//				arg is the ptr to in\out buffer
	//======================================================================

	u_char			*buf;
	PIOCTL_DATA 	pIoctlData;
	BYTE			byStatus 	= 0;
	int 			nCount;
	int 			nRetVal		= 0;
	DWORD 			dwLen;
	PUCHAR			pucTemp;
	ULONG			ulSentLen,
					ulOffset;
	DEVICE_EXTENSION	*pDeviceExtension = NULL;
	// temporary variable
	size_t lTempCount;

	buf = (u_char *) arg;

	DebugLog (" IFDHIoctl Entry: %u for device [%d]\n", ucIoctl, iminor (inode));

	pDeviceExtension = &sDeviceExtension [ iminor (inode) ];

	if( NULL == pDeviceExtension )
	{
		DebugLog (" IFDHIoctl : pDeviceExtension is NULL!\n");
		return -ENODEV;
	}

	if ( pDeviceExtension->nStop )
	{
		//	Added by Vijay. Device has been removed
		return -ENODEV;
	}

 	DebugLog (" IFDHIoctl : pDeviceExtension is NOT NULL!\n");

	pIoctlData = (PIOCTL_DATA) kmalloc (sizeof (IOCTL_DATA), GFP_KERNEL);
	if(!pIoctlData)
	{
		DebugLog (" IFDHIoctl : No memory for IOCTL Data");
		return -ENODEV;
	}

 	DebugLog ("IFDHIoctl : Allocated memory for pIOCTLDATA\n");

	if(pDeviceExtension->irqFlag == TRUE)
	{
		pDeviceExtension->irqFlag = FALSE;
// 		run_task_queue(&tq_acr92);
	}

	pucTemp = (PUCHAR) kmalloc(sizeof(UCHAR)*2 , GFP_KERNEL);
	if( NULL == pucTemp)
	{
		DebugLog("IFDHIoctl : Error Allocating Memory\n");
		return -ENODEV;
	}

 	DebugLog ("IFDHIoctl : Allocated memory for pucTemp\n");

// 	DebugLog ("\n IFDHIoctl : Inside IFDHIoctl with IOCTL : %u\n", ucIoctl);

	lTempCount = copy_from_user (pIoctlData, (PIOCTL_DATA) buf, sizeof (IOCTL_DATA));

	switch ( ucIoctl )
	{
		case CMD_CARD_STATUS:
			pIoctlData->ucRxBuffer[0] = pDeviceExtension->byCardStatus;
			pIoctlData->dwRxLength = 1;
// 			pIoctlData->dwRxLength = 2;
			byStatus = 2;
			nRetVal = 2;
			break;

		case CMD_BOOTROM_STATUS:
			pIoctlData->ucRxBuffer[0] = pDeviceExtension->bBootRomPresent;
			pIoctlData->dwRxLength = 1;
			DebugLog (" IFDHIoctl : Got bootrom status as %X\n", pIoctlData->ucRxBuffer[0]);
			byStatus = 2;
			nRetVal = 2;
			break;

		case CMD_PCCARD_DOWNLOAD_START:
			byStatus = IFDPCCardDownloadStart (
				pDeviceExtension , buf );
			if( STATUS_SUCCESS != byStatus )
			{
				byStatus = -ENODEV;
			}
			pDeviceExtension->pucFwBuffer = (PUCHAR)
				kmalloc(sizeof(UCHAR)*65536 , GFP_KERNEL);

			if( NULL ==  pDeviceExtension->pucFwBuffer )
			{
				byStatus = -ENODEV;
				break;
			}
			memset( pDeviceExtension->pucFwBuffer, 0, 65536 );
			pDeviceExtension->ulOffset = 0;

			nRetVal = 1;
			break;

		case CMD_PCCARD_DOWNLOAD_DATA:
			ulSentLen = pIoctlData->dwTxLength;
			ulOffset = pDeviceExtension->ulOffset;
			lTempCount = copy_from_user(
				(PUCHAR)(pDeviceExtension->pucFwBuffer+ulOffset ),
				(PUCHAR) pIoctlData->ucRxBuffer,
				(ULONG) ulSentLen
				);

			memcpy( (PUCHAR)pDeviceExtension->pucFwBuffer+ulOffset, (PUCHAR)pIoctlData->ucTxBuffer,ulSentLen);
//			for(nCount=0;nCount< ulSentLen; nCount++ )
//			{
//				//DebugLog(" <(%02x)>, ", (pDeviceExtension->pucFwBuffer[ulOffset+nCount] ) );
//			}
			pDeviceExtension->ulOffset += ulSentLen;
			nRetVal = 1;
			break;

		case CMD_PCCARD_DOWNLOAD_END:
			DebugLog (" IFDHIoctl : Before IFDPerformPCCardDownLoad\n");
			byStatus = IFDPerformPCCardDownLoad(pDeviceExtension, pDeviceExtension->pucFwBuffer);
			DebugLog (" IFDHIoctl : After IFDPerformPCCardDownLoad\n");
			if( STATUS_SUCCESS != byStatus )
			{
				byStatus = -ENODEV;
				break;
			}
			nRetVal = 1;
			pIoctlData->dwRxLength = 2;
			break;

		case CMD_PCCARD_ERASE:
			IFDPerformPCCardErase( pDeviceExtension );
			break;

		case CMD_WRITE_DATA:
			DebugLog( " Writing to Device \n");
			if( TRUE == pDeviceExtension->bBootRomPresent )
			{
				break;
			}
			nRetVal = LocalWrite(pDeviceExtension, pIoctlData->ucTxBuffer, pIoctlData->dwTxLength );
			break;

		case CMD_READ_DATA:
			DebugLog( " Reading from Device \n");
			if( TRUE == pDeviceExtension->bBootRomPresent )
			{
				break;
			}
 			dwLen = LocalRead(pDeviceExtension, pIoctlData->ucRxBuffer, pIoctlData->dwRxLength );

			if( dwLen > 0 )
			{
				byStatus = dwLen;
				nRetVal = (int) dwLen;
			}

			if ( pIoctlData->dwRxLength < 0 )
			{
				byStatus = dwLen;
				nRetVal = -1;
				break;
			}

			pIoctlData->dwRxLength = dwLen;
			DebugLog(" Reading Data of Length [%ld] \n",	pIoctlData->dwRxLength );

			for(nCount=0; nCount < pIoctlData->dwRxLength; nCount++ )
			{
				DebugLog(" [%02X] ", pIoctlData->ucRxBuffer[nCount] );
			}
			break;

		case CMD_POLLED_MODE:
			DebugLog(" IFDHIOCTL :: CMD_POLLED_MODE is %02X \n",pIoctlData->ucRxBuffer[0]);

			if( 0x01 == pIoctlData->ucRxBuffer[0] )
			{
				if(FALSE == pDeviceExtension->bIsPolledMode )
				{
					pDeviceExtension->bIsPolledMode = TRUE;
					init_waitqueue_head(&pDeviceExtension->tqsCardStatusWait);
					pDeviceExtension->DE_thread_pid = kernel_thread(PcdmCardStatusThread, pDeviceExtension, CLONE_KERNEL);
					DebugLog (" IFDHIoctl : ********** Thread created  **********\n");
				}
			}
			else
			{
				if( TRUE == pDeviceExtension->bIsPolledMode )
				{
					DebugLog(" Killing Polled Mode Card Tracking \n");
					pDeviceExtension->bIsPolledMode = FALSE;
					pDeviceExtension->bIsIrqSet = pDeviceExtension->bOrgIrqStatus;
					send_sig(SIGTERM, pDeviceExtension->pDE_thread, 1);
					pDeviceExtension->pDE_thread = 0;
				}
			}
			break;

		case CMD_GET_IRQSTATUS:
			pIoctlData->ucRxBuffer[0] = pDeviceExtension->bIsIrqSet;
			pIoctlData->dwRxLength = 1;
			DebugLog (" IFDHIoctl : Get IRQ status -> %X\n", pIoctlData->ucRxBuffer[0]);
			break;

		case CMD_SET_REVISION:
			pDeviceExtension->byPcdmRev = pIoctlData->ucRxBuffer[0];
			DebugLog("Setting PCDM Revision [%x] \n",pIoctlData->ucRxBuffer[0]);
			break;

		default:
			DebugLog("<0> WRONG IOCTL being sent %d \n", ucIoctl );
			byStatus = -ENODEV;
			nRetVal = -ENODEV;
			break;
	}

	pDeviceExtension->inIOCTL = FALSE;

	if(pIoctlData)
	{
		if(buf)
		{
			lTempCount = copy_to_user (buf, (u_char *) pIoctlData, sizeof (IOCTL_DATA));
		}

		kfree(pIoctlData);
		pIoctlData = NULL;
	}

	if( pucTemp )
	{
		kfree(pucTemp);
		pucTemp = NULL;
	}

	DebugLog (" IFDHIoctl EXIT\n");

	return nRetVal; //TBD  to return appropriate values.
}

ssize_t IFDHRead(
	struct file 	*file,
	char 		*pucRxBuffer,
	size_t 		count,
	loff_t 		*loc
	)
{
	DEVICE_EXTENSION	*pDeviceExtension = NULL;
	PIOCTL_DATA pIoctlData = (PIOCTL_DATA) pucRxBuffer;
 	BYTE byTemp = READER_STS_CARD_ABSENT;
	DWORD dwSize = 1;
	size_t lTempCount;

//	DebugLog ("\n Entering IFDHRead! \n");

	pDeviceExtension = &sDeviceExtension[ iminor (file->f_dentry->d_inode) ];

	if( !pucRxBuffer )
	{
		DebugLog (" IFDHRead : pucRxBuffer NULL\n");
	}

	lTempCount = copy_to_user (&pIoctlData->dwRxLength, &dwSize, sizeof(dwSize));

//	DebugLog (" IFDHRead Card Status : %x\n", pDeviceExtension->byCardStatus);

	if( pDeviceExtension )
	{
		if ( pDeviceExtension->nStop )
		{
			// Added by Vijay. Device has been removed.
			DebugLog (" Device removed...\n");
			lTempCount = copy_to_user (pIoctlData->ucRxBuffer, &byTemp, 1);
			return 0;
		}

		if( TRUE == pDeviceExtension->bIsSystemResume )
		{
			lTempCount = copy_to_user (pIoctlData->ucRxBuffer, &byTemp, 1);
			pDeviceExtension->bIsSystemResume = FALSE;
		}
		else
		{
 			lTempCount = copy_to_user (pIoctlData->ucRxBuffer, &pDeviceExtension->byCardStatus, 1);
 //			DebugLog (" IFDHRead Card Status : %d\n", pDeviceExtension->byCardStatus);
		}
	}
	else
	{
 		DebugLog (" pDeviceExtension is NULL!\n");
		lTempCount = copy_to_user (pIoctlData->ucRxBuffer, &byTemp, 1);
 	}

//	DebugLog(" IFDHRead EXIT\n");

	return 1;
}

DWORD LocalRead( DEVICE_EXTENSION *pDeviceExtension, char *pucRxBuffer, size_t count )
{
	/*======================================================================

	ioctl calls from the ifdhandler.
	cmd is the IOCTL code
	arg is the ptr to in\out buffer

	======================================================================*/
	PUCHAR pucData = NULL;
	NTSTATUS NTStatus = STATUS_SUCCESS;
	DWORD dwDataLen;
	DWORD dwRetVal;
	//int i;

//	DebugLog ("\n\tLocalRead Entry\n");

	pucData = (PUCHAR) kmalloc((sizeof(UCHAR) * 2048), GFP_KERNEL);

	if( NULL == pucData )
	{
		return -ENODEV; // TBD
	}
	memset ( pucData, 0, ( sizeof(UCHAR)*2048) );

	NTStatus = PcdmRead (pDeviceExtension, pucData, &dwDataLen);

	if( STATUS_SUCCESS == NTStatus )
	{
// 		for(i=0;i < dwDataLen; i++ )
// 		{
// 			DebugLog(" LocalRead::Before Copying, Recieved Values %02X \n", *(pucData+i));
// 		}

		memcpy ( pucRxBuffer, pucData, dwDataLen);

//		for(i=0;i < dwDataLen; i++ )
//		{
//			DebugLog(" (%02X)", *(pucRxBuffer+i));
//		}
//		DebugLog(" \n");

		dwRetVal = dwDataLen;
	}
	else
	{
		dwRetVal = 0;
	}

	if( pucData )
	{
		kfree( pucData );
		pucData = NULL;
	}

//	DebugLog ("\tLocalRead EXIT");

	return dwRetVal;
}

ssize_t IFDHWrite(
	struct file *file,
	const char *pucTxBuffer,
	size_t szCount,
	loff_t *loc)
{
	DebugLog (" IFDHWrite called!\n");
	return -ENODEV;
}

ssize_t LocalWrite( DEVICE_EXTENSION *pDeviceExtension, const char *pucTxBuffer, size_t szCount)
{
	PUCHAR 		pucData = NULL;
	NTSTATUS 	NTStatus = STATUS_SUCCESS;
	DWORD 		dwDataLen = 0;
	int 		nI;

	DebugLog ("\n\tLocalWrite Entry\n");

 	pucData = (PUCHAR) kmalloc( (sizeof(UCHAR) * 2048) , GFP_KERNEL );
 	if( NULL == pucData )
 	{
 		return -ENODEV;
 	}
 	memset ( pucData, 0, ( sizeof(UCHAR)*2048) );

	dwDataLen = (DWORD)szCount;
	memcpy ( pucData, pucTxBuffer, dwDataLen );

	DebugLog("Length of data being sent down from IFDHWrite is %d", szCount);
	for(nI=0;nI<(int)szCount;nI++)
	{
		DebugLog("[%02X]", *(pucData+nI));
	}
	DebugLog("\n");

	NTStatus = PcdmWrite (pDeviceExtension, pucData, dwDataLen);

	if( pucData )
	{
		kfree( pucData );
		pucData = NULL;
	}

	if( STATUS_SUCCESS == NTStatus )
	{
		return szCount;
	}
	else
	{
		return 0;
	}

	DebugLog ("\tLocalWrite EXIT\n");
}

irqreturn_t Driver_IrqHandler(int irq,void *pDev,struct pt_regs *regs)
{
	/*====================================================================
	This is the Interrupt Handler
	Interrupts occur during Smart Card Insertion and removal.
	Hence card status is updated here.
	IFDGetCardStatus() sends a command to the reader, and command can be sent
	only if there is no other command in progress in the reader.
	If a command is in progress, the task is queued
	=====================================================================*/
	DEVICE_EXTENSION *pDeviceExtension;
	pDeviceExtension = (PDEVICE_EXTENSION) pDev;

 	DebugLog( "\n Calling Driver_IrqHandler \n" );

	if ( pDeviceExtension->nStop )
	{
		return IRQ_NONE;
	}

	// we dont use the interrupt handler if forced polled
	// mode is set in the registry.
	if( TRUE == pDeviceExtension->bIsPolledMode )
	{
		DebugLog(" Polling is set. Hence, not handled here\n");
		return IRQ_NONE;
	}

 	PcmciaIrqServiceRoutine ( pDeviceExtension );

// 	DebugLog( " Driver_IrqHandler EXIT\n" );

	return IRQ_HANDLED;
}

void printdummy(char *dbg_msg, ...)
{
	// This is a dummy function to override the debug data
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
void tqGetCardStatus( void *arg)
#else
void tqGetCardStatus( struct work_struct *pstWorkQueue)
#endif
{
	/*==========================================================
	This is a scheduler queue used to trigger the card status
	function if the interrupt got is from our device
	===========================================================*/
	
	PDEVICE_EXTENSION pDeviceExtension;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
	pDeviceExtension = (PDEVICE_EXTENSION) arg;
#else
	pDeviceExtension = container_of (pstWorkQueue, DEVICE_EXTENSION, stWorkQueue );
#endif

	DebugLog ("\n tqGetCardStatus Entry\n");

	if ( pDeviceExtension->nStop )
	{
		return;
	}

	if( FALSE == pDeviceExtension->bBootRomPresent )
	{
		DebugLog (" tqGetCardStatus : IFDGetCardStatus\n");
		IFDGetCardStatus( pDeviceExtension );
	}
	pDeviceExtension->bIsQueueSet = FALSE;

	DebugLog (" tqGetCardStatus EXIT\n");
}

int PcdmCardStatusThread(void *arg)
{
	BYTE                    byStatus;
	BYTE                    byReg6;
	PPCDM_REGISTERS         IoBase;

	PDEVICE_EXTENSION pDeviceExtension = (PDEVICE_EXTENSION) arg;

	IoBase = (PPCDM_REGISTERS) pDeviceExtension->IoBase;

	DebugLog("\n Entering CardStatus Thread \n");

	if( TRUE == pDeviceExtension->bBootRomPresent )
	{
		DebugLog("  Issue IN PcdmCardStatusThread 1 \n" );
	}
	// whatever be the state, we clear this register.
	outb( 0x00, (unsigned long) &IoBase->DummyReg6 );

	allow_signal(SIGTERM);

	while(!signal_pending(current))
	{
		if ( NULL == pDeviceExtension )
		{
			DebugLog("NUll pointer in DeviceExtension. breaking from thread loop \n");
			break;
		}

// 		 fall asleep for 200 milli Seconds
		interruptible_sleep_on_timeout(&pDeviceExtension->tqsCardStatusWait, (HZ/2) );

// 		We need to do a memory barrier here to be sure that
// 		the flags are visible on all CPUs.
		mb();

// 		here we are back from sleep, either due to the timeout
// 		(one second), or because we caught a signal.

		if ( NULL == pDeviceExtension )
		{
			DebugLog("NUll pointer in DeviceExtension. breaking from thread loop \n");
			break;
		}

		if( FALSE == pDeviceExtension->bBootRomPresent )
		{
			byStatus = inb( (unsigned long) &IoBase->CmdStatusReg );
			byReg6 = inb ( (unsigned long) &IoBase->DummyReg6 );
			if (FALSE == ( byStatus & STS_ERR ) )
			{
				if( 0x01 == byReg6 )
				{
					DebugLog(" Polled Mode::Card Status Change \n");
					if( TRUE == pDeviceExtension->bBootRomPresent )
					{
						DebugLog("Issue in PcdmCardStatusThread 2 \n" );
					}
					outb( 0x00, (unsigned long) &IoBase->DummyReg6 );

					DebugLog (" PcdmCardStatusThread : IFDGetCardStatus\n");
					IFDGetCardStatus( pDeviceExtension );
				}
			}
			else
			{
				break;
			}
		}
	}

	if(signal_pending(current))
	{
		// we received a request to terminate ourself
		DebugLog("  Recieved Termination Request. Breaking from loop\n" );
	}

	DebugLog("\n CardStatus Thread EXIT\n");

	return 0;
}

module_init( init_driver );
module_exit( exit_driver );

MODULE_LICENSE("GPL");
